%%% File    : erlog.erl
%%% Author  : Robert Virding
%%% Purpose : Main interface to the Erlog interpreter.
%%%
%%% (C)Robert Virding. This stuff is mine, distributed without
%%% warranty "as is" and may not be used commercially without written
%%% permission.
%%%
%%% Structures	- {Functor,arg1, Arg2,...} where Functor is an atom
%%% Variables	- {Name} where Name is an atom or integer
%%% Lists	- Erlang lists
%%% Atomic	- Erlang constants
%%%
%%% There is no problem with the representation of variables as Prolog
%%% functors of arity 0 are atoms. This representation is much easier
%%% to test for, and create new variables with than using funny atom
%%% names like '$1' (yuch!), and we need LOTS of variables.

-module(erlog).

%% Shell functions.
-export([shell/0,show_bindings/1]).
%% Interface to server.
-export([prove/2,next_solution/1,consult/2,reconsult/2,
	 get_db/1,set_db/2,halt/1]).
-export([start/0,start_link/0,return_bindings/1]).
%% User utilities.
-export([is_legal_term/1,vars_in/1]).

-import(lists, [foldl/3,foreach/2]).

%%-compile(export_all).

%% prove(Erlog, Goal) -> {succeed,Bindings} | fail.
%% next_solution(Erlog) -> {succeed,Bindings} | fail.
%% consult(Erlog, File) -> ok | {error,Error}.
%% reconsult(Erlog, File) -> ok | {error,Error}.
%% get_db(Erlog) -> {ok,Database}.
%% set_db(Erlog, Database) -> ok.
%% halt(Erlog) -> ok.
%% Interface functions to server.

prove(Erl, Goal) ->
    send_request(Erl, {prove,Goal}),
    wait_reply().

next_solution(Erl) ->
    send_request(Erl, next_solution),
    wait_reply().

consult(Erl, File) ->
    send_request(Erl, {consult,File}),
    wait_reply().

reconsult(Erl, File) ->
    send_request(Erl, {reconsult,File}),
    wait_reply().

get_db(Erl) ->
    send_request(Erl, get_db),
    wait_reply().

set_db(Erl, Db) ->
    send_request(Erl, {set_db,Db}),
    wait_reply().

halt(Erl) ->
    send_request(Erl, halt),
    wait_reply().

%% start()
%% start_link()
%% Start an Erlog server.

start() ->
    spawn(fun () -> server() end).

start_link() ->
    spawn_link(fun () -> server() end).

server() ->
    Db = erlog_int:built_in_db(),
    server_loop(Db).

server_loop(Db0) ->
    receive
	{erlog_request,From,Req} ->
	    io:format("R: ~p\n", [{erlog_request,From,Req}]),
	    request(From, Req, Db0)
    end.	    

request(From, {prove,Goal}, Db0) ->
    prove_server_goal(From, Goal, Db0);
request(From, next_solution, Db0) ->
    send_reply(From, fail),
    server_loop(Db0);
request(From, {consult,File}, Db0) ->
    case consult_file(File, Db0) of
	{ok,Db1} ->
	    send_reply(From, ok),
	    server_loop(Db1);
	{erlog_error,Error} ->
	    send_reply(From, {error,Error}),
	    server_loop(Db0);
	{error,Error} ->
	    send_reply(From, {error,Error}),
	    server_loop(Db0)
    end;
request(From, {reconsult,File}, Db0) ->
    case reconsult_file(File, Db0) of
	{ok,Db1} ->
	    %%send_reply(From, ok),
	    send_reply(From, Db1),
	    server_loop(Db1);
	{erlog_error,Error} ->
	    send_reply(From, {error,Error}),
	    server_loop(Db0);
	{error,Error} ->
	    send_reply(From, {error,Error}),
	    server_loop(Db0)
    end;
request(From, get_db, Db) ->
    send_reply(From, {ok,Db}),
    server_loop(Db);
request(From, {set_db,Db1}, Db0) ->
    send_reply(From, ok),
    server_loop(Db1);
request(From, halt, Db0) ->
    send_reply(From, ok).

%% prove_server_goal(From, Goal, Database)
%% Try to prove Goal sent to us by From. Use the process dictionary to
%% pass the From and Request from return_bindings when it receives a
%% new request. Caveat: all solutions to a request are sent to the
%% process which requested the goal, not the senders of the
%% next_solutions.

prove_server_goal(From, Goal, Db0) ->
    Vs = vars_in(Goal),
    put(return_bindings, From),
    %% Adding the show_bindings call in this fashion means that the
    %% variables will be dereferenced by the intepreter. Must use
    %% 'catch' here as 'try' does not do last-call optimisation.
    case catch erlog_int:prove_goal({',',Goal,
				     {ecall,{':',erlog,
					     {return_bindings,Vs}},[]}},
				    Db0) of
	{succeed,Bs,Db1} ->
	    %% This is an unrecognised request to return_bindings.
	    {NewFrom,NewReq} = erase(return_bindings),
	    request(NewFrom, NewReq, Db1);
	{fail,Db1} ->
	    send_reply(erase(return_bindings), fail),
	    server_loop(Db1);
	%% Errors from the Erlog interpreters.
	{erlog_error,Error,Db1} ->
	    send_reply(erase(return_bindings), {error,Error}),
	    server_loop(Db1);
	{erlog_error,Error} ->		%No new database
	    send_reply(erase(return_bindings), {error,Error}),
	    server_loop(Db0);
	%% Errors and exits from user code.
	{'EXIT',Error} ->
	    send_reply(erase(return_bindings), {'EXIT',Error}),
	    server_loop(Db0)
    end.

%% return_bindings(VarList) -> {succeed_last,[]} | fail.
%% Show the bindings and query user for next solution. Called through
%% ecall so variables will be dereferenced. Use the process dictionary
%% to pass the From and new Request up to prove_server_goal. Naughty,
%% naughty!

return_bindings(Vs) ->
    %% foreach(fun ({Name,Val}) ->
    %%		    io:format("_~w = ~p\n", [Name,Val]) end,
    %%	    Vs),
    %% Tell them we proved the goal.
    send_reply(erase(return_bindings), {succeed,Vs}),
    receive
	{erlog_request,From,next_solution} ->
	    put(return_bindings, From),
	    fail;
	{erlog_request,From,Req} ->
	    put(return_bindings, {From,Req}),	%Need to pass this up!
	    {succeed_last,[]}
    end.

%% send_request(Erlog, Request)
%% send_reply(To, Reply)
%% wait_reply() -> Reply.

send_request(Erl, Req) -> Erl ! {erlog_request,self(),Req}.
    
send_reply(To, Rep) -> To ! {erlog_reply,Rep}.

wait_reply() ->
    receive
	{erlog_reply,Rep} -> Rep
    end.

%% shell()
%% A simple Erlog shell similar to a "normal" Prolog shell. It allows
%% user to enter goals, see resulting bindings and request next
%% solution.

shell() ->
    Db = erlog_int:built_in_db(),
    shell_loop(Db).

shell_loop(Db0) ->
    Line = io:get_line('| ?- '),
    case erl_scan:string(Line) of
	{ok,Ts,_} ->
	    case erlog_parse:term(Ts) of
		{ok,halt} -> ok;		%Break out of the loop
		{ok,Files} when is_list(Files) ->
		    case reconsult_files(Files, Db0) of
			{ok,Db1} ->
			    io:format("Yes\n"),
			    shell_loop(Db1);
			{erlog_error,Error} ->
			    io:format("Error: ~p\n", [Error]),
			    shell_loop(Db0);
			{error,{_,Pm,Pe}} ->
			    io:format("Error: ~s\n", [Pm:format_error(Pe)]),
			    shell_loop(Db0);
			{error,Error} ->
			    io:format("Error: ~p\n", [Error]),
			    shell_loop(Db0)
		    end;
		{ok,Goal} ->
		    prove_shell_goal(Goal, Db0);
		{error,{_,Pm,Pe}} ->
		    io:format("Error: ~s\n", [Pm:format_error(Pe)]),
		    shell_loop(Db0)
	    end;
	{error,{_,Sm,Se}} ->
	    io:format("Error: ~s\n", [Sm:format_error(Se)]),
	    shell_loop(Db0)
    end.

reconsult_files([F|Fs], Db0) ->
    case reconsult_file(F, Db0) of
	{ok,Db1} -> reconsult_files(Fs, Db1);
	{erlog_error,Error} -> {erlog_error,Error};
	{error,Error} -> {error,Error}
    end;
reconsult_files([], Db) -> {ok,Db};
reconsult_files(Other, _Db) -> {error,{type_error,list,Other}}.

prove_shell_goal(Goal, Db0) ->
    Vs = vars_in(Goal),
    %% Adding the show_bindings call in this fashion means that the
    %% variables will be dereferenced by the intepreter. Must use
    %% 'catch' here as 'try' does not do last-call optimisation.
    case catch erlog_int:prove_goal({',',Goal,
				     {ecall,{':',erlog,{show_bindings,Vs}},[]}},
				    Db0) of
	{succeed,Bs,Db1} ->
	    io:format("Yes\n"),
	    shell_loop(Db1);
	{fail,Db1} ->
	    io:format("No\n"),
	    shell_loop(Db1);
	%% Errors from the Erlog interpreters.
	{erlog_error,Error,Db1} ->
	    io:format("Error: ~p\n", [Error]),
	    shell_loop(Db1);
	{erlog_error,Error} ->	%No new database here
	    io:format("Error: ~p\n", [Error]),
	    shell_loop(Db0);
	%% Errors and exits from user code.
	{'EXIT',Error} ->	%No new database here
	    io:format("EXIT: ~p\n", [Error]),
	    shell_loop(Db0)
    end.

%% show_bindings(VarList) -> {succeed_last,Val} | fail.
%% Show the bindings and query user for next solution. Called through
%% ecall so variables will be dereferenced.

show_bindings([]) -> {succeed_last,[]};		%Just once here
show_bindings(Vs) ->
    foreach(fun ({Name,Val}) ->
		    io:format("~w = ~p\n", [Name,Val]) end,
	    Vs),
    Line = io:get_line(': '),
    case string:chr(Line, $;) of
	0 -> {succeed_last,[]};			%No ";", done
	_ -> fail				%Found ";", another solution
    end.

%% vars_in(Term) -> [{Name,Var}].
%% Returns an ordered list of {VarName,Variable} pairs.

vars_in(Term) -> vars_in(Term, orddict:new()).

vars_in({'_'}, Vs) -> Vs;			%Never in!
vars_in({Name}=Var, Vs) -> orddict:store(Name, Var, Vs);
vars_in(Struct, Vs) when tuple(Struct) ->
    vars_in_struct(Struct, 2, size(Struct), Vs);
vars_in([H|T], Vs) ->
    vars_in(T, vars_in(H, Vs));
vars_in(_Atomic, Vs) -> Vs.

vars_in_struct(Str, I, S, Vs) when I > S -> Vs;
vars_in_struct(Str, I, S, Vs) ->
    vars_in_struct(Str, I+1, S, vars_in(element(I, Str), Vs)).

%% consult_file(File, Database) ->
%%	{ok,NewDatabase} | {error,Error} | {erlog_error,Error}.
%% reconsult_file(File, Database) ->
%%	{ok,NewDatabase} | {error,Error} | {erlog_error,Error}.
%% Load/reload an Erlog file into the interpreter. Reloading will
%% abolish old definitons of clauses.

consult_file(File, Db0) ->
    case load_file(File) of
	{ok,Terms} ->
	    consult_terms(fun (Term, Db) ->
				  {ok,erlog_int:assertz_clause(Term, Db)}
			  end, Db0, Terms);
	Error -> Error
    end.

reconsult_file(File, Db0) ->
    case load_file(File) of
	{ok,Terms} ->
	    Ifun = fun (Term, {Db0,Seen}) ->
			   Func = functor(Term),
			   case lists:member(Func, Seen) of
			       true ->
				   {ok,{erlog_int:assertz_clause(Term, Db0),
					Seen}};
			       false ->
				   Db1 = erlog_int:abolish_clauses(Func, Db0),
				   {ok,{erlog_int:assertz_clause(Term, Db1),
					[Func|Seen]}}
			   end
		   end,
	    case consult_terms(Ifun, {Db0,[]}, Terms) of
		{ok,{Db1,Seen1}} -> {ok,Db1};
		Error -> Error
	    end;
	Error -> Error
    end.

consult_terms(Ifun, Db0, [T|Ts]) ->
    case catch Ifun(T, Db0) of
	{ok,Db1} -> consult_terms(Ifun, Db1, Ts);
	{erlog_error,E,_Db1} -> {erlog_error,E}
    end;
consult_terms(Ifun, Db, []) -> {ok,Db}.

functor({':-',H,_B}) -> erlog_int:functor(H);
functor(T) -> erlog_int:functor(T).

%% load_file(FileName) -> {ok,[Term]} | {error,Error}.
%% Load a file containing Prolog terms. This has been taken from io
%% but cleaned up using try.

load_file(File) ->
    case file:open(File, [read]) of
	{ok,Fd} ->
	    try
		{ok,load_stream(io:scan_erl_exprs(Fd, ''), Fd)}
	    catch
		throw:Term -> Term;
		error:_Error -> {error,einval};
		exit:_Error -> {error,einval}
	    after
		file:close(Fd)
	    end;
	Error -> Error
    end.

load_stream({ok,Toks,Line}, Fd) ->
    case erlog_parse:term(Toks) of
	{ok,Term} ->
	    [Term|load_stream(io:scan_erl_exprs(Fd, '', Line), Fd)];
	{error,What} -> throw({error,What})
    end;
load_stream({error,What}, _Fd) ->
    throw({error, What});
load_stream({eof,_}, _Fd) -> [].

%% is_legal_term(Goal) -> true | false.
%% Test if a goal is a legal Erlog term. Basically just check if
%% tuples are used correctly as structures and variables.

is_legal_term({V}) -> is_atom(V);
is_legal_term([H|T]) ->
    is_legal_term(H) andalso is_legal_term(T);
is_legal_term(T) when is_tuple(T) ->
    if  size(T) >= 2, is_atom(element(1, T)) ->
	    are_legal_args(T, 2, size(T));	%The right tuples.
	true -> false
    end;
is_legal_term(T) when is_constant(T) -> true;	%All constants, including []
is_legal_term(_T) -> false.

are_legal_args(T, I, S) when I > S -> true;
are_legal_args(T, I, S) ->
    is_legal_term(element(I, T)) andalso are_legal_args(T, I+1, S).
