%%% File    : erlog_int.erl
%%% Author  : Robert Virding
%%% Purpose : Basic interpreter of a Prolog sub-set.
%%%
%%% (C)Robert Virding. This stuff is mine, distributed without
%%% warranty "as is" and may not be used commercially without written
%%% permission.
%%% 
%%% This is the basic Prolog interpreter. 
%%% The internal data structures used are very direct and basic:
%%%
%%% Structures	- {Functor,arg1, Arg2,...} where Functor is an atom
%%% Variables	- {Name} where Name is an atom or integer
%%% Lists	- Erlang lists
%%% Atomic	- Erlang constants
%%%
%%% There is no problem with the representation of variables as Prolog
%%% functors of arity 0 are atoms. This representation is much easier
%%% to test for, and create new variables with than using funny atom
%%% names like '$1' (yuch!), and we need LOTS of variables.
%%%
%%% The (internal) functions for proving goals/bodies/calls all have
%%% the same return values:
%%%
%%% {succeed,NewBindings,NewVarNum,NewDatabase} - the goal succeeded
%%%	and these are the new bindings/varnum/database to continue with.
%%%
%%% {fail,NewDatabase} - the goal failed and this is the new database.
%%%
%%% When the interpreter detects an error it builds an error term
%%%
%%%	{erlog_error,ErrorDescriptor,Database}
%%%
%%% and throws it. The ErrorDescriptor is a valid Erlog term.

-module(erlog_int).

%% Main interface functions.
-export([prove_goal/2]).
-export([built_in_db/0]).
%% Utilities.
-export([new_bindings/0,deref/2,deref_all/2,functor/1]).
-export([asserta_clause/2,assertz_clause/2,abolish_clauses/2]).

-import(lists, [map/2,foldl/3]).

%%-compile(export_all).

%% built_in_db() -> Database.
%% Create an initial clause database containing the built-in
%% predicates and predefined library predicates.

built_in_db() ->
    Db0 = new_db(),
    %% First add the Erlang built-ins.
    Db1 = foldl(fun (Head, Db) -> add_built_in(Head, Db) end, Db0,
		[
		 %% Logic and control.
		 {call,{1}},
		 {',',{1},{2}},
		 '!',
		 {';',{1},{2}},
		 fail,
		 {'->',{1},{2}},
		 {'\\+',{1}},
		 {once,{1}},
		 repeat,
		 true,
		 %% Term creation and decomposition.
		 {arg,{1},{2},{3}},
		 {copy_term,{1},{2}},
		 {functor,{1},{2},{3}},
		 {'=..',{1},{2}},
		 %% Arithmetic evaluation and comparison
		 {'>',{1},{2}},
		 {'>=',{1},{2}},
		 {'=:=',{1},{2}},
		 {'=\\=',{1},{2}},
		 {'<',{1},{2}},
		 {'=<',{1},{2}},
		 {'is',{1},{2}},
		 %% Clause creation and destruction.
		 {abolish,{1}},
		 {assert,{1}},
		 {asserta,{1}},
		 {assertz,{1}},
		 {retract,{1}},
		 {retractall,{1}},
		 %% Clause retrieval and information.
		 {clause,{1},{2}},
		 {current_predicate,{1}},
		 {predicate_property,{1},{2}},
		 %% Type testing.
		 {atom,{1}},
		 {atomic,{1}},
		 {compound,{1}},
		 {integer,{1}},
		 {float,{1}},
		 {number,{1}},
		 {nonvar,{1}},
		 {var,{1}},
		 %% All solutions
		 %% Term unification and comparison
		 {'@>',{1},{2}},
		 {'@>=',{1},{2}},
		 {'==',{1},{2}},
		 {'\\==',{1},{2}},
		 {'@<',{1},{2}},
		 {'@=<',{1},{2}},
		 {'=',{1},{2}},
		 {'\\=',{1},{2}},
		 %% External interface
		 {ecall,{1},{2}}]),
    %% Next the interpreted built-ins.
    Db2 = foldl(fun (Clause, Db) -> assertz_clause(Clause, Db) end, Db1,
 		[
 		]),
    %% Compiled built-ins and common libray.
    Db3 = foldl(fun ({Head,M,F}, Db) ->
			add_compiled_proc(Head, M, F, Db) end, Db2,
		[
		 %% {{app,{1},{2},{3}},user_pl,app_3},
		 %% {{rev,{1},{2}},user_pl,rev_2}
		]),
    %% Finally some interpreted common library.
    foldl(fun (Clause, Db) -> assertz_clause(Clause, Db) end, Db3,
	  [
	   {'C',[{1}|{2}],{1},{2}},		%For DCGs, someday
	   {append,[],{1},{1}},
	   {':-',{append,[{1}|{2}],{3},[{1}|{4}]},{append,{2},{3},{4}}},
	   {insert,{1},{2},[{2}|{1}]},
	   {':-',{insert,[{1}|{2}],{3},[{1}|{4}]},{insert,{2},{3},{4}}},
	   {member,{1},[{1}|{2}]},
	   {':-',{member,{1},[{2}|{3}]},{member,{1},{3}}}
	   ]).

%% prove_goal(Goal, Database) ->
%%	{succeed,Bindings,NewDatabase} | {fail,NewDatabase}.
%% This is the main entry point into the interpreter. Check that
%% everything is consistent then prove the goal as a call.

prove_goal(Goal0, Db0) ->
    %% Check term and build new instance of term with bindings.
    {Goal1,Bs0,Vn0} = initial_goal(Goal0),
    Goal2 = check_goal(Goal1, Bs0, Db0),	%Check we have a proper goal
    case prove_body(Goal2, Bs0, Vn0, Db0) of
	{succeed,Bs,Vn1,Db1} -> {succeed,Bs,Db1};
	Fail -> Fail
    end.

%% prove_goal(Goal, NextGoal, Bindings, VarNum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Prove one goal. We seldom return succeed here but usually go directly to
%% to NextGoal.
%% Handle built-in predicates here. RTFM for a description of the
%% built-ins. Hopefully we do the same.

%% Logic and control.
prove_goal({call,G0}, Next, Bs, Vn, Db) ->
    G1 = check_goal(G0, Bs, Db),
    prove_goal(G1, Next, Bs, Vn, Db);
prove_goal({',',L,R}, Next, Bs, Vn, Db) ->
    prove_goal(L, {',',R,Next}, Bs, Vn, Db);
prove_goal({';',{'->',I,T},E}, Next, Bs0, Vn0, Db0) ->
    %% This has already been checked earlier by our creator.
    %% Can never fail back into I.
    case prove_body(I, Bs0, Vn0, Db0) of
	{succeed,Bs1,Vn1,Db1} -> prove_goal(T, Next, Bs1, Vn1, Db1);
	{fail,Db1} -> prove_goal(E, Next, Bs0, Vn0, Db1)
    end;
prove_goal({';',L,R}, Next, Bs, Vn, Db0) ->
    case prove_goal(L, Next, Bs, Vn, Db0) of
	{fail,Db1} -> prove_goal(R, Next, Bs, Vn, Db1);
	Succeed -> Succeed
    end;
prove_goal({'->',I,T}, Next, Bs0, Vn0, Db0) ->
    %% This has already been checked earlier by our creator.
    %% Can never fail back into I.
    case prove_body(I, Bs0, Vn0, Db0) of
	{succeed,Bs1,Vn1,Db1} -> prove_goal(T, Next, Bs1, Vn1, Db1);
	Fail -> Fail
    end;
prove_goal(fail, _Next, _Bs, _Vn, Db) -> {fail,Db};
prove_goal({'\\+',G0}, Next, Bs0, Vn0, Db0) ->
    %% Can never fail back into G.
    G1 = check_goal(G0, Bs0, Db0),
    case prove_body(G1, Bs0, Vn0, Db0) of
	{fail,Db1} -> prove_body(Next, Bs0, Vn0, Db1);
	{succeed,Bs1,Vn1,Db1} -> {fail,Db1}
    end;
prove_goal({once,G0}, Next, Bs0, Vn0, Db0) ->
    %% Can never fail back into G.
    G1 = check_goal(G0, Bs0, Db0),
    case prove_body(G1, Bs0, Vn0, Db0) of
	{succeed,Bs1,Vn1,Db1} -> prove_body(Next, Bs1, Vn1, Db1);
	Fail -> Fail
    end;
prove_goal(repeat, Next, Bs, Vn, Db0) ->
    case prove_body(Next, Bs, Vn, Db0) of
	{fail,Db1} -> prove_goal(repeat, Next, Bs, Vn, Db1);
	Succeed -> Succeed
    end;
prove_goal(true, Next, Bs, Vn, Db) ->
    prove_body(Next, Bs, Vn, Db);
%% Term unification and comparison
prove_goal({'@>',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('>', L, R, Next, Bs, Vn, Db);
prove_goal({'@>=',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('>=', L, R, Next, Bs, Vn, Db);
prove_goal({'==',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('==', L, R, Next, Bs, Vn, Db);
prove_goal({'\\==',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('/=', L, R, Next, Bs, Vn, Db);
prove_goal({'@<',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('<', L, R, Next, Bs, Vn, Db);
prove_goal({'@=<',L,R}, Next, Bs, Vn, Db) ->
    term_test_prove_body('=<', L, R, Next, Bs, Vn, Db);
prove_goal({'=',L,R}, Next, Bs, Vn, Db) ->
    unify_prove_body(L, R, Next, Bs, Vn, Db);
prove_goal({'\\=',L,R}, Next, Bs0, Vn, Db) ->
    case unify(L, R, Bs0) of
	{succeed,Bs1} -> {fail,Db};
	fail -> prove_body(Next, Bs0, Vn, Db)
    end;
%% Type testing.
prove_goal({atom,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	T when is_atom(T) -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
prove_goal({atomic,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	[] -> prove_body(Next, Bs, Vn, Db);
	T when is_constant(T) -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
prove_goal({compound,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	[] -> {fail,Db};
	T when is_constant(T) -> {fail,Db};
	Other -> prove_body(Next, Bs, Vn, Db)
    end;
prove_goal({integer,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	T when is_integer(T) -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
prove_goal({float,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	T when is_float(T) -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
prove_goal({number,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	T when is_number(T) -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
prove_goal({nonvar,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	{T} -> {fail,Db};
	Other -> prove_body(Next, Bs, Vn, Db)
    end;
prove_goal({var,T0}, Next, Bs, Vn, Db) ->
    case deref(T0, Bs) of
	{T} -> prove_body(Next, Bs, Vn, Db);
	Other -> {fail,Db}
    end;
%% Arithmetic evalution and comparison.
prove_goal({'>',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('>', L, R, Next, Bs, Vn, Db);
prove_goal({'>=',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('>=', L, R, Next, Bs, Vn, Db);
prove_goal({'=:=',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('==', L, R, Next, Bs, Vn, Db);
prove_goal({'=\\=',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('/=', L, R, Next, Bs, Vn, Db);
prove_goal({'<',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('<', L, R, Next, Bs, Vn, Db);
prove_goal({'=<',L,R}, Next, Bs, Vn, Db) ->
    arith_test_prove_body('=<', L, R, Next, Bs, Vn, Db);
prove_goal({is,N,E0}, Next, Bs, Vn, Db) ->
    E = eval_arith(deref(E0, Bs), Bs, Db),
    unify_prove_body(N, E, Next, Bs, Vn, Db);
%% Term creation and decomposition.
prove_goal({arg,I0,Ct0,A0}, Next, Bs0, Vn, Db) ->
    case {deref(I0, Bs0),deref(Ct0, Bs0)} of
	{I,[H|T]} ->
	    if  I == 1 -> unify_prove_body(H, A0, Next, Bs0, Vn, Db);
		I == 2 -> unify_prove_body(T, A0, Next, Bs0, Vn, Db);
		true -> {fail,Db}
	    end;
	{I,Ct} when is_integer(I), is_tuple(Ct), size(Ct) >= 2 ->
	    if  I > 1, I + 1 =< size(Ct) ->
		    unify_prove_body(element(I+1, Ct), A0, Next, Bs0, Vn, Db);
		true -> {fail,Db}
	    end;
	%%Type failure just generates an error.
	{I,Ct} when not(is_integer(I)) -> type_error(integer, I, Db);
	{I,Ct} -> type_error(compound, Ct, Db)
    end;
prove_goal({copy_term,T0,C}, Next, Bs, Vn0, Db) ->
    %% Use new_instance to create the copy, can ignore orddict it creates.
    {T,_Nbs,Vn1} = new_instance(deref_all(T0, Bs), Vn0),
    unify_prove_body(T, C, Next, Bs, Vn1, Db);
prove_goal({functor,T0,F0,A0}, Next, Bs0, Vn0, Db) ->
    case deref_all(T0, Bs0) of
	T when is_tuple(T), size(T) >= 2 ->
	    unify_prove_body(F0, element(1, T),
			     A0, size(T)-1, Next, Bs0, Vn0, Db);
	T when is_constant(T) ; T == [] ->
	    unify_prove_body(F0, T, A0, 0, Next, Bs0, Vn0, Db);
	[_|_] ->				%He, he, he!
	    unify_prove_body(F0, '.', A0, 2, Next, Bs0, Vn0, Db);
	{_}=Var ->
	    case {deref_all(F0, Bs0),deref_all(A0, Bs0)} of
		{'.',2} ->			%He, he, he!
		    Bs1 = add_binding(Var, [{Vn0}|{Vn0+1}], Bs0),
		    prove_body(Next, Bs1, Vn0+2, Db);
		{F,0} when is_constant(F) ->
		    Bs1 = add_binding(Var, F, Bs0),
		    prove_body(Next, Bs1, Vn0, Db);
		{F,A} when is_atom(F), is_integer(A), A > 0 ->
		    As = make_vars(A, Vn0),
		    Bs1 = add_binding(Var, list_to_tuple([F|As]), Bs0),
		    prove_body(Next, Bs1, Vn0+A, Db); %!!!
		%% Now the error cases.
		{{_},A} -> instantiation_error(Db);
		{F,A} when is_atom(F) -> type_error(integer, A, Db);
		{F,A} -> type_error(atom, F, Db)
	    end
    end;
prove_goal({'=..',T0,L0}, Next, Bs0, Vn0, Db) ->
    case deref_all(T0, Bs0) of
	T when is_tuple(T), size(T) >= 2 ->
	    Es = tuple_to_list(T),
	    unify_prove_body(L0, Es, Next, Bs0, Vn0, Db);
	T when is_constant(T) ; T == [] ->
	    unify_prove_body(L0, [T], Next, Bs0, Vn0, Db);
	[Lh|Lt] ->				%He, he, he!
	    unify_prove_body(L0, ['.',Lh,Lt], Next, Bs0, Vn0, Db);
	{_}=Var ->
	    case deref_all(L0, Bs0) of
		['.',Lh,Lt] ->			%He, he, he!
		    Bs1 = add_binding(Var, [Lh|Lt], Bs0),
		    prove_body(Next, Bs1, Vn0, Db);
		[A] when is_constant(A) ->
		    Bs1 = add_binding(Var, A, Bs0),
		    prove_body(Next, Bs1, Vn0, Db);
		[F|As] when is_atom(F), length(As) > 0 ->
		    Bs1 = add_binding(Var, list_to_tuple([F|As]), Bs0),
		    prove_body(Next, Bs1, Vn0, Db);
		%% Now the error cases.
		[{_}|_] -> instantiation_error(Db);
		Other -> type_error(list, Other, Db)
	    end
    end;
%% Clause creation and destruction.
prove_goal({abolish,Pi0}, Next, Bs, Vn, Db) ->
    case deref_all(Pi0, Bs) of
	{'/',N,A} when atom(N), integer(A), A > 0 ->
	    prove_body(Next, Bs, Vn, abolish_clauses({N,A}, Db));
	Pi -> type_error(predicate_indicator, Pi, Db)
    end;
prove_goal({assert,C0}, Next, Bs, Vn, Db) ->
    C = deref_all(C0, Bs),
    prove_body(Next, Bs, Vn, assertz_clause(C, Db));
prove_goal({asserta,C0}, Next, Bs, Vn, Db) ->
    C = deref_all(C0, Bs),
    prove_body(Next, Bs, Vn, asserta_clause(C, Db));
prove_goal({assertz,C0}, Next, Bs, Vn, Db) ->
    C = deref_all(C0, Bs),
    prove_body(Next, Bs, Vn, assertz_clause(C, Db));
prove_goal({retract,C0}, Next, Bs, Vn, Db) ->
    C = deref_all(C0, Bs),
    prove_retract(C, Next, Bs, Vn, Db);
%% Clause retrieval and information
prove_goal({clause,H0,B}, Next, Bs, Vn, Db) ->
    H1 = deref_all(H0, Bs),
    prove_clause(H1, B, Next, Bs, Vn, Db);
prove_goal({current_predicate,Pi0}, Next, Bs, Vn, Db) ->
    Pi = deref_all(Pi0, Bs),
    prove_current_predicate(Pi, Next, Bs, Vn, Db);
prove_goal({predicate_property,H0,P}, Next, Bs, Vn, Db) ->
    H = deref_all(H0, Bs),
    case catch get_procedure_type(functor(H), Db) of
	built_in -> unify_prove_body(P, built_in, Next, Bs, Vn, Db);
	compiled -> unify_prove_body(P, compiled, Next, Bs, Vn, Db);
	interpreted -> unify_prove_body(P, interpreted, Next, Bs, Vn, Db);
	undefined -> {fail,Db};
	{erlog_error,E} -> erlog_error(E, Db)
    end;	
%% External interface
prove_goal({ecall,C0,Val}, Next, Bs, Vn, Db) ->
    %% Build the initial call.
    Efun = case deref_all(C0, Bs) of
	       {':',M,F} when is_atom(M), is_atom(F) ->
		   fun () -> M:F() end;
	       {':',M,{F,A}} when is_atom(M), is_atom(F) ->
		   fun () -> M:F(A) end;
	       {':',M,{F,A1,A2}} when is_atom(M), is_atom(F) ->
		   fun () -> M:F(A1,A2) end;
	       {':',M,T} when is_atom(M), is_tuple(T), size(T) >= 2,
			      is_atom(element(1, T)) ->
			  L = tuple_to_list(T),
		   fun () -> apply(M, hd(L), tl(L)) end;
	       Fun when is_function(Fun) -> Fun;
	       Other -> type_error(callable, Other, Db)
	   end,
    prove_ecall(Efun, Val, Next, Bs, Vn, Db);
%% Now look up the database.
prove_goal(G, Next, Bs, Vn, Db) ->
    %% io:fwrite("PG: ~p\n    ~p\n    ~p\n", [deref_all(G, Bs),Next,Bs]),
    case catch get_procedure(functor(G), Db) of
	{code,{Mod,Func}} -> Mod:Func(G, Next, Bs, Vn, Db);
	{clauses,Cs} -> prove_goal_clauses(G, Cs, Next, Bs, Vn, Db);
	undefined -> {fail,Db};
	%% Getting built_in here is an error!
	{erlog_error,E} -> erlog_error(E, Db)	%Fill in more error data
    end.

%% check_goal(Goal, Bindings, Database) -> WellFormedGoall.
%% Check to see that Goal is bound and ensure that it is well-formed.

check_goal(G0, Bs, Db) ->
    case deref_all(G0, Bs) of
	{V} ->  instantiation_error(Db);	%Must have something to call
	G1 ->
	    case catch well_form_body(G1) of
		{erlog_error,E} -> erlog_error(E, Db);
		G2 -> G2
	    end
    end.

%% unify_prove_body(Term1, Term2, Next, Bindings, VarNum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Unify Term1 = Term2, on success prove body Next else fail.

unify_prove_body(T1, T2, Next, Bs0, Vn, Db) ->
    case unify(T1, T2, Bs0) of
	{succeed,Bs1} -> prove_body(Next, Bs1, Vn, Db);
	fail -> {fail,Db}
    end.

%% unify_prove_body(A1, B1, A2, B2, Next, Bindings, VarNum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Unify A1 = B1, A2 = B2, on success prove body Next else fail.

unify_prove_body(A1, B1, A2, B2, Next, Bs0, Vn, Db) ->
    case unify(A1, B1, Bs0) of
	{succeed,Bs1} -> unify_prove_body(A2, B2, Next, Bs1, Vn, Db);
	fail -> {fail,Db}
    end.

%% term_test_prove_body(Test, Left, Right, Next, Bindings, Varnum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.

term_test_prove_body(Test, L, R, Next, Bs, Vn, Db) ->
    case erlang:Test(deref_all(L, Bs), deref_all(R, Bs)) of
	true -> prove_body(Next, Bs, Vn, Db);
	false -> {fail,Db}
    end.


%% arith_test_prove_body(Test, Left, Right, Next, Bindings, VarNum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.

arith_test_prove_body(Test, L, R, Next, Bs, Vn, Db) ->
    case erlang:Test(eval_arith(deref(L, Bs), Bs, Db),
		     eval_arith(deref(R, Bs), Bs, Db)) of
	true -> prove_body(Next, Bs, Vn, Db);
	false -> {fail,Db}
    end.

%% prove_ecall(Generator, Value, Next, Bindings, VarNum, Database) ->
%%	{succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Call an external (Erlang) generator and handle return value, either
%% succeed or fail.

prove_ecall(Efun, Val, Next, Bs0, Vn, Db0) ->
    case Efun() of
	{succeed,Ret,Cont} ->			%Succeed and more choices
	    case unify(Val, Ret, Bs0) of
		{succeed,Bs1} ->
		    case prove_body(Next, Bs1, Vn, Db0) of
			{fail,Db1} ->
			    prove_ecall(Cont, Val, Next, Bs0, Vn, Db1);
			Succeed -> Succeed
		    end;
		fail -> prove_ecall(Cont, Val, Next, Bs0, Vn, Db0)
	    end;
	{succeed_last,Ret} ->			%Succeed but last choice
	    unify_prove_body(Val, Ret, Next, Bs0, Vn, Db0);
	fail -> {fail,Db0}			%No more
    end.

%% prove_clause(Head, Body, Next, Bindings, VarNum, DataBase) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Unify clauses matching with functor from Head with both Head and Body.

prove_clause(H, B, Next, Bs, Vn, Db) ->
    Functor = functor(H),
    case get_procedure(Functor, Db) of
	{clauses,Cs} -> unify_clauses(H, B, Cs, Next, Bs, Vn, Db);
	{code,_} -> permission_error(pred_ind(Functor), Db);
	built_in -> permission_error(pred_ind(Functor), Db);
	undefined -> {fail,Db}
    end.

%% unify_clauses(Head, Body, Clauses, Next, Bindings, VarNum, Database) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Try to unify Head and Body using Clauses which all have the same functor.

unify_clauses(Ch, Cb, [C], Next, Bs0, Vn0, Db) ->
    %% No choice point on last clause
    case unify_clause(Ch, Cb, C, Bs0, Vn0) of
	{succeed,Bs1,Vn1} -> prove_body(Next, Bs1, Vn1, Db);
	fail -> {fail,Db}
    end;
unify_clauses(Ch, Cb, [C|Cs], Next, Bs0, Vn0, Db0) ->
    case unify_clause(Ch, Cb, C, Bs0, Vn0) of
	{succeed,Bs1,Vn1} ->
	    case prove_body(Next, Bs1, Vn1, Db0) of
		{succeed,Bs2,Vn2,Db2} -> {succeed,Bs2,Vn2,Db2};
		{fail,Db2} -> unify_clauses(Ch, Cb, Cs, Next, Bs0, Vn0, Db2)
	    end;
	fail -> unify_clauses(Ch, Cb, Cs, Next, Bs0, Vn0, Db0)
    end;
unify_clauses(_Ch, _Cb, [], _Next, _Bs, _Vn, Db) -> {fail,Db}.

unify_clause(Ch, Cb, {_Tag,H0,B0}, Bs0, Vn0) ->
    {H1,Rs1,Vn1} = new_instance(H0, Vn0),	%Unique vars on head first
    case unify(Ch, H1, Bs0) of
	{succeed,Bs1} ->
	    {B1,Rs2,Vn2} = new_instance(B0, Rs1, Vn1),	%Now we need the rest
	    case unify(Cb, B1, Bs1) of
		{succeed,Bs2} -> {succeed,Bs2,Vn2};
		fail -> fail
	    end;
	fail -> fail
    end.

%% prove_current_predicate(PredInd, Next, Bindings, VarNum, DataBase) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Match functors of existing user (interpreted) predicate with PredInd.

prove_current_predicate(Pi, Next, Bs, Vn, Db) ->
    case Pi of
	{'/',F,A} -> ok;
	{_} -> ok;
	Other -> type_error(predicate_indicator, Other)
    end,
    Fs = get_interp_functors(Db),
    prove_predicates(Pi, Fs, Next, Bs, Vn, Db).

prove_predicates(Pi, [F|Fs], Next, Bs0, Vn0, Db0) ->
    case unify(Pi, pred_ind(F), Bs0) of
	{succeed,Bs1} ->
	    case prove_body(Next, Bs1, Vn0, Db0) of
		{fail,Db1} -> prove_predicates(Pi, Fs, Next, Bs0, Vn0, Db0);
		Succeed -> Succeed
	    end;
	fail -> prove_predicates(Pi, Fs, Next, Bs0, Vn0, Db0)
    end;
prove_predicates(Pi, [], Next, Bs, Vn, Db) -> {fail,Db}.

%% prove_goal_clauses(Goal, Clauses, Next, Bindings, VarNum, Database) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Try to prove Goal using Clauses which all have the same functor.

prove_goal_clauses(G, [C], Next, Bs, Vn, Db) ->	%No choice point on last clause
    prove_goal_clause(G, C, Next, Bs, Vn, Db);
prove_goal_clauses(G, [C|Cs], Next, Bs, Vn, Db0) ->
    case prove_goal_clause(G, C, Next, Bs, Vn, Db0) of
	{fail,Db1} -> prove_goal_clauses(G, Cs, Next, Bs, Vn, Db1);
	Succeed -> Succeed
    end;
prove_goal_clauses(_G, [], _Next, _Bs, _Vn, Db) -> {fail,Db}.

prove_goal_clause(G, {_Tag,H0,B0}, Next, Bs0, Vn0, Db) ->
    %%io:fwrite("PGC: ~p\n", [{_Tag,H0,B0}]),
    {H,Rs1,Vn1} = new_instance(H0, Vn0),	%Unique vars on head first
    case unify(G, H, Bs0) of
	{succeed,Bs1} ->
	    {B,Rs2,Vn2} = new_instance(B0, Rs1, Vn1),	%Now we need the rest
	    prove_body(prepend_body(B, Next), Bs1, Vn2, Db);
	fail -> {fail,Db}
    end.

prepend_body({',',L,R}, Next) -> {',',L,prepend_body(R, Next)};
prepend_body(true, Next) -> Next;
prepend_body(G, Next) -> {',',G,Next}.

%% prove_retract(Clause, Next, Bindings, VarNum, Database) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Retract clauses in database matching Clause.

prove_retract({':-',H,B}, Next, Bs, Vn, Db) ->
    prove_retract(H, B, Next, Bs, Vn, Db);
prove_retract(H, Next, Bs, Vn, Db) ->
    prove_retract(H, true, Next, Bs, Vn, Db).

prove_retract(H, B, Next, Bs, Vn, Db) ->
    Functor = functor(H),
    case get_procedure(Functor, Db) of
	{clauses,Cs} -> retract_clauses(H, B, Cs, Next, Bs, Vn, Db);
	{code,_} -> permission_error(pred_ind(Functor), Db);
	built_in -> permission_error(pred_ind(Functor), Db);
	undefined -> {fail,Db}
    end.

%% retract_clauses(Head, Body, Clauses, Next, Bindings, VarNum, Database) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Try to retract Head and Body using Clauses which all have the same functor.

retract_clauses(Ch, Cb, [C|Cs], Next, Bs0, Vn0, Db0) ->
    case unify_clause(Ch, Cb, C, Bs0, Vn0) of
	{succeed,Bs1,Vn1} ->
	    %% We have found a right clause so now retract it.
	    Db1 = retract_clause(functor(Ch), element(1, C), Db0),
	    case prove_body(Next, Bs1, Vn1, Db1) of
		{succeed,Bs2,Vn2,Db2} -> {succeed,Bs2,Vn2,Db2};
		{fail,Db2} ->
		    retract_clauses(Ch, Cb, Cs, Next, Bs0, Vn0, Db2)
	    end;
	fail -> retract_clauses(Ch, Cb, Cs, Next, Bs0, Vn0, Db0)
    end;
retract_clauses(_Ch, _Cb, [], _Next, _Bs, _Vn, Db) -> {fail,Db}.

%% prove_body(Body, Bindings, VarNum, Database) ->
%%     {succeed,NewBindings,NewVarNum,NewDatabase} | {fail,NewDatabase}.
%% Recognise special body constructs first.

prove_body(true, Bs, Vn, Db) -> {succeed,Bs,Vn,Db};	%No more body
prove_body({',',L,R}, Bs0, Vn0, Db0) ->
    case prove_goal(L, R, Bs0, Vn0, Db0) of
	{succeed,Bs1,Vn1,Db1} -> {succeed,Bs1,Vn1,Db1};
	{fail,Db1} -> {fail,Db1}
    end;
prove_body(G, Bs, Vn, Db) -> prove_goal(G, true, Bs, Vn, Db).

%% new_instance(Term, VarNum) -> {Term,NewRepls,NewVarNum}.
%% new_instance(Term, Repls, VarNum) -> {Term,NewRepls,NewVarNum}.
%%  Generate a new instance of a term with new, fresh unused
%%  variables. No bindings from original variables to new
%%  variables. It can handle replacing integer variables with
%%  overlapping integer ranges. Don't check Term as it should already
%%  be checked. Use orddict as there will seldom be many variables and
%%  it it fast to setup.

new_instance(A, Vn) -> new_instance(A, orddict:new(), Vn).

new_instance([], Rs, Vn) -> {[],Rs,Vn};
new_instance([H0|T0], Rs0, Vn0) ->
    {H,Rs1,Vn1} = new_instance(H0, Rs0, Vn0),
    {T,Rs2,Vn2} = new_instance(T0, Rs1, Vn1),
    {[H|T],Rs2,Vn2};
new_instance({'_'}, Rs, Vn) -> {{Vn},Rs,Vn+1};	%Unique variable
new_instance({V0}, Rs0, Vn0) ->			%Other variables
    case orddict:find(V0, Rs0) of
	{ok,V1} -> {{V1},Rs0,Vn0};
	error -> {{Vn0},orddict:store(V0, Vn0, Rs0),Vn0+1}
    end;
new_instance(T0, Rs0, Vn0) when tuple(T0) ->
    Es0 = tuple_to_list(T0),
    {Es1,Rs1,Vn1} = new_instance(Es0, Rs0, Vn0),
    {list_to_tuple(Es1),Rs1,Vn1};
new_instance(A, Rs, Vn) -> {A,Rs,Vn}.		%Constant

%% unify(Term, Term, Bindings) -> {succeed,NewBindings} | fail.
%% Unify two terms with a set of bindings.

unify(T10, T20, Bs0) ->
    case {deref(T10, Bs0),deref(T20, Bs0)} of
	{T1,T2} when constant(T1), constant(T2), T1 == T2 -> {succeed,Bs0};
	{{V},{V}} -> {succeed,Bs0};
	{{_}=Var,T2} -> {succeed,add_binding(Var, T2, Bs0)};
	{T1,{_}=Var} -> {succeed,add_binding(Var, T1, Bs0)};
	{[H1|T1],[H2|T2]} ->
	    case unify(H1, H2, Bs0) of
		{succeed,Bs1} -> unify(T1, T2, Bs1);
		fail -> fail
	    end;
	{[],[]} -> {succeed,Bs0};
	{T1,T2} when tuple(T1), tuple(T2), size(T1) == size(T2),
		     element(1, T1) == element(1, T2) ->
	    unify_args(T1, T2, Bs0, 2, size(T1));
	_Other -> fail
    end.

unify_args(S1, S2, Bs, I, S) when I > S -> {succeed,Bs};
unify_args(S1, S2, Bs0, I, S) ->
    case unify(element(I, S1), element(I, S2), Bs0) of
	{succeed,Bs1} -> unify_args(S1, S2, Bs1, I+1, S);
	fail -> fail
    end.

%% eval_arith(ArithExpr, Bindings, Database) -> Number.
%% Evaluate an arithmetic expression, include the database for errors.

eval_arith({'+',A,B}, Bs, Db) ->
    eval_arith(deref(A, Bs), Bs, Db) + eval_arith(deref(B, Bs), Bs, Db);
eval_arith({'-',A,B}, Bs, Db) ->
    eval_arith(deref(A, Bs), Bs, Db) - eval_arith(deref(B, Bs), Bs, Db);
eval_arith({'*',A,B}, Bs, Db) ->
    eval_arith(deref(A, Bs), Bs, Db) * eval_arith(deref(B, Bs), Bs, Db);
eval_arith({'/',A,B}, Bs, Db) ->
    eval_arith(deref(A, Bs), Bs, Db) / eval_arith(deref(B, Bs), Bs, Db);
eval_arith({'**',A,B}, Bs, Db) ->
    math:pow(eval_arith(deref(A, Bs), Bs, Db),
	     eval_arith(deref(B, Bs), Bs, Db));
eval_arith({'//',A,B}, Bs, Db) ->
    eval_int(deref(A, Bs), Bs, Db) div eval_int(deref(B, Bs), Bs, Db);
eval_arith({'rem',A,B}, Bs, Db) ->
    eval_int(deref(A, Bs), Bs, Db) rem eval_int(deref(B, Bs), Bs, Db);
eval_arith({'+',A}, Bs, Db) ->
    + eval_arith(deref(A, Bs), Bs, Db);
eval_arith({'-',A}, Bs, Db) ->
    - eval_arith(deref(A, Bs), Bs, Db);
eval_arith({'abs',A}, Bs, Db) ->
    abs(eval_arith(deref(A, Bs), Bs, Db));
eval_arith({'float',A}, Bs, Db) ->
    float(eval_arith(deref(A, Bs), Bs, Db));
eval_arith({'truncate',A}, Bs, Db) ->
    trunc(eval_arith(deref(A, Bs), Bs, Db));
eval_arith(N, Bs, Db) when number(N) -> N;	%Just a number
%% Error cases.
eval_arith({N}, Bs, Db) -> instantiation_error(Db);
eval_arith(N, Bs, Db) when tuple(N) ->
    type_error(evaluable, pred_ind(element(1, N), size(N)-1), Db);
eval_arith([_|_], Bs, Db) -> type_error(evaluable, pred_ind('.', 2), Db);
eval_arith(O, Bs, Db) -> type_error(evaluable, O, Db).

%% eval_int(IntegerExpr, Bindings, Database) -> Integer.
%% Evaluate an integer expression, include the database for errors.

eval_int(E0, Bs, Db) ->
    E = eval_arith(E0, Bs, Db),
    if  integer(E) -> E;
	true -> type_error(integer, E, Db)
    end.

%% make_vars(Count, VarNum) -> [Var].
%% Make a list of new variables starting at VarNum.

make_vars(0, Vn) -> [];
make_vars(I, Vn) ->
    [{Vn}|make_vars(I-1, Vn+1)].

%% Errors

type_error(Type, Value, Db) -> erlog_error({type_error,Type,Value}, Db).
type_error(Type, Value) -> erlog_error({type_error,Type,Value}).

instantiation_error(Db) -> erlog_error(instantiation_error, Db).

permission_error(Value, Db) -> erlog_error({permission_error,Value}, Db).

erlog_error(E, Db) -> throw({erlog_error,E,Db}).
erlog_error(E) -> throw({erlog_error,E}).

%% Database
%% The database is a dict where the key is the functor pair {Name,Arity}.
%% The value is: built_in | {NextTag,[{Tag,Head,Clause}]}.
%% Built-ins are defined by the system and cannot manipulated by user
%% code.
%% We are a little paranoid here and do our best to ensure consistency
%% in the database by checking input arguments even if we know they
%% come from "good" code.

%%-define(DB, orddict).
-define(DB, dict).

new_db() -> ?DB:new().

%% add_built_in(Head, Database) -> NewDatabase.
%% Add Head as a built-in in the database.

add_built_in(H, Db) ->
    Functor = functor(H),
    ?DB:store(Functor, built_in, Db).

%% add_compiled_proc(Head, Module, Function, Database) -> NewDatabase.
%% Add Head as a compiled procedure with code in Module:Function. No
%% checking.

add_compiled_proc(H, M, F, Db) ->
    Functor = functor(H),
    ?DB:store(Functor, {code,{M,F}}, Db).

%% assertz_clause(Clause, Database) -> NewDatabase.
%% assertz_clause(Head, Body, Database) -> NewDatabase.
%% Assert a clause into the database first checking that it is well
%% formed.

assertz_clause({':-',H,B}, Db) -> assertz_clause(H, B, Db);
assertz_clause(H, Db) -> assertz_clause(H, true, Db).

assertz_clause(Head, Body0, Db) ->
    {Functor,Body} = case catch {functor(Head), well_form_body(Body0)} of
			 {erlog_error,E} -> erlog_error(E, Db);
			 FB -> FB
		      end,
    ?DB:update(Functor,
	       fun (built_in) -> permission_error(pred_ind(Functor), Db);
		   ({code,_}) -> permission_error(pred_ind(Functor), Db);
		   ({clauses,T,Cs}) -> {clauses,T+1,Cs ++ [{T,Head,Body}]}
	       end, {clauses,1,[{0,Head,Body}]}, Db).

%% asserta_clause(Clause, Database) -> NewDatabase.
%% asserta_clause(Head, Body, Database) -> NewDatabase.
%% Assert a clause into the database first checking that it is well
%% formed.

asserta_clause({':-',H,B}, Db) -> asserta_clause(H, B, Db);
asserta_clause(H, Db) -> asserta_clause(H, true, Db).

asserta_clause(Head, Body0, Db) ->
    {Functor,Body} = case catch {ok,functor(Head),well_form_body(Body0)} of
			 {erlog_error,E} -> erlog_error(E, Db);
			 {ok,F,B} -> {F,B}
		      end,
    ?DB:update(Functor,
	       fun (built_in) -> permission_error(pred_ind(Functor), Db);
		   ({code,_}) -> permission_error(pred_ind(Functor), Db);
		   ({clauses,T,Cs}) -> {clauses,T+1,[{T,Head,Body}|Cs]}
	       end, {clauses,1,[{0,Head,Body}]}, Db).

%% retract_clause(Functor, ClauseTag, Database) -> NewDatabase.
%% Retract (remove) the clause with tag ClauseTag from the list of
%% clauses of Functor.

retract_clause(F, Ct, Db) ->
    case ?DB:find(F, Db) of
	{ok,built_in} -> permission_error(F, Db);
	{ok,{code,_}} -> permission_error(F, Db);
	{ok,{clauses,Nt,Cs}} ->
	    ?DB:store(F, {clauses,Nt,lists:keydelete(Ct, 1, Cs)}, Db);
	error -> Db				%Do nothing
    end.

%% abolish_clauses(Functor, Database) -> NewDatabase.

abolish_clauses(Func, Db) ->
    case ?DB:find(Func, Db) of
	{ok,built_in} -> permission_error(Func, Db);
	{ok,{compiled,_}} -> ?DB:erase(Func, Db);
	{ok,{clauses,_,_}} -> ?DB:erase(Func, Db);
	error -> Db				%Do nothing
    end.

%% get_procedure(Functor, Database) ->
%%	built_in | {compiled,{Mod,Func}} | {interpreted,[Clause]} | undefined.
%% Return the procedure type and data for a functor.

get_procedure(Func, Db) ->
    case ?DB:find(Func, Db) of
	{ok,built_in} -> built_in;		%A built-in
	{ok,{code,{M,F}}=P} -> P;		%Compiled (perhaps someday)
	{ok,{clauses,_T,Cs}} -> {clauses,Cs};	%Interpreted clauses
	error -> undefined			%Undefined
    end.

%% get_procedure_type(Functor, Database) ->
%%	built_in | compiled | interpreted | undefined.
%% Return the procedure type for a functor.

get_procedure_type(Func, Db) ->
    case ?DB:find(Func, Db) of
	{ok,built_in} -> built_in;		%A built-in
	{ok,{code,_}} -> compiled;		%Compiled (perhaps someday)
	{ok,{clauses,_,_}} -> interpreted;	%Interpreted clauses
	error -> undefined			%Undefined
    end.

%% get_interp_functors(Database) -> [Functor].

get_interp_functors(Db) ->
    ?DB:fold(fun (Func, built_in, Fs) -> Fs;
		 (Func, {code,_}, Fs) -> Fs;
		 (Func, {clauses,_,_}, Fs) -> [Func|Fs]
	     end, [], Db).

functor(T) when tuple(T), size(T) > 1, atom(element(1, T)) ->
    {element(1, T),size(T)-1};
functor(T) when atom(T) -> {T,0};
functor(T) -> type_error(callable, T).

%% well_form_body(Body) -> Body.
%% Check that Body is well-formed, flatten conjunctions and add
%% explicit call to top-level variables.

well_form_body({',',L,R}) -> well_form_conj(L, R);
well_form_body({';',L,R}) ->
    {';',well_form_body(L),well_form_body(R)};
well_form_body({'->',I,T}) ->
    {'->',well_form_body(I),well_form_body(T)};
well_form_body({V}) -> {call,{V}};		%Top-level vars become call
well_form_body(Goal) -> functor(Goal), Goal.	%Check goal

well_form_conj({',',L,R}, More) ->
    well_form_conj(L, well_form_conj(R, More));
well_form_conj(B, More) ->
    {',',well_form_body(B),well_form_body(More)}.

pred_ind({N,A}) -> {'/',N,A}.

pred_ind(N, A) -> {'/',N,A}.

%% Bindings
%% Bindings are kept in a dict where the key is the variable name.

%%-define(BIND, orddict).
-define(BIND, dict).

new_bindings() -> ?BIND:new().

add_binding({V}, Val, Bs0) ->
    ?BIND:store(V, Val, Bs0).

get_binding({V}, Bs) ->
    ?BIND:find(V, Bs).

deref({V}=T0, Bs) ->
    case ?BIND:find(V, Bs) of
	{ok,T1} -> deref(T1, Bs);
	error -> T0
    end;
deref(T, Bs) -> T.				%Not a variable, return it.

deref_all(A, Rs) when constant(A) -> A;
deref_all([], Rs) -> [];
deref_all([H0|T0], Rs) ->
    [deref_all(H0, Rs)|deref_all(T0, Rs)];
deref_all({V}=Var, Rs) ->
    case ?BIND:find(V, Rs) of
	{ok,T} -> deref_all(T, Rs);
	error -> Var
    end;
deref_all(T, Rs) when tuple(T) ->
    Es0 = tuple_to_list(T),
    Es1 = deref_all(Es0, Rs),
    list_to_tuple(Es1).

%% well_form_term(Goal, VarNum) -> {Goal,NewVarNum}.
%% Check goal for well-formedness as an Erlog term and replace '_'
%% variables with unique numbered variables. Error on non-well-formed
%% goals.

well_form_term({'_'}, Vn) -> {{Vn},Vn+1};
well_form_term({Name}=Var, Vn) when is_atom(Name) ->
    {Var,Vn};
well_form_term([H0|T0], Vn0) ->
    {H1,Vn1} = well_form_term(H0, Vn0),
    {T1,Vn2} = well_form_term(T0, Vn1),
    {[H1|T1],Vn2};
well_form_term([], Vn) -> {[],Vn};
well_form_term(S, Vn0)
  when is_tuple(S), size(S) >= 2, is_atom(element(1, S)) ->
    As0 = tl(tuple_to_list(S)),
    {As1,Vn1} = well_form_term(As0, Vn0),
    {list_to_tuple([element(1, S)|As1]),Vn1};
well_form_term(T, Vn) when is_constant(T) -> {T,Vn};
well_form_term(T, Vn) -> type_error(callable, T).

%% initial_goal(Goal) -> {Goal,NewBindings,NewVarNum}.
%% initial_goal(Goal, Bindings, VarNum) -> {Goal,NewBindings,NewVarNum}.
%% Check term for well-formedness as an Erlog term and replace '_'
%% variables with unique numbered variables. Error on non-well-formed
%% goals.

initial_goal(Goal) -> initial_goal(Goal, new_bindings(), 0).

initial_goal({'_'}, Bs, Vn) -> {{Vn},Bs,Vn+1};	%Anonymous variable
initial_goal({Name}=Var0, Bs, Vn) when is_atom(Name) ->
    case get_binding(Var0, Bs) of
	{ok,Var1} -> {Var1,Bs,Vn};
	error ->
	    Var1 = {Vn},
	    {Var1,add_binding(Var0, Var1, Bs),Vn+1}
    end;
initial_goal([H0|T0], Bs0, Vn0) ->
    {H1,Bs1,Vn1} = initial_goal(H0, Bs0, Vn0),
    {T1,Bs2,Vn2} = initial_goal(T0, Bs1, Vn1),
    {[H1|T1],Bs2,Vn2};
initial_goal([], Bs, Vn) -> {[],Bs,Vn};
initial_goal(S, Bs0, Vn0)
  when is_tuple(S), size(S) >= 2, is_atom(element(1, S)) ->
    As0 = tl(tuple_to_list(S)),
    {As1,Bs1,Vn1} = initial_goal(As0, Bs0, Vn0),
    {list_to_tuple([element(1, S)|As1]),Bs1,Vn1};
initial_goal(T, Bs, Vn) when is_constant(T) -> {T,Bs,Vn};
initial_goal(T, _Bs, _Vn) -> type_error(callable, T).
